// SPDX-License-Identifier: Unlicense
pragma solidity ^0.8.24;

import "./Reputation.sol";
import "./Forum.sol";
import "../interfaces/IOnValidate.sol";
import "hardhat/console.sol";

struct ValidationPoolStake {
    uint id;
    bool inFavor;
    uint amount;
    address sender;
}

struct ValidationPoolParams {
    uint duration;
    uint[2] quorum; // [ Numerator, Denominator ]
    uint[2] winRatio; // [ Numerator, Denominator ]
    uint bindingPercent;
    bool redistributeLosingStakes;
}

struct ValidationPool {
    uint id;
    uint postIndex;
    address sender;
    uint minted;
    mapping(uint => ValidationPoolStake) stakes;
    uint stakeCount;
    ValidationPoolParams params;
    uint fee;
    uint endTime;
    bool resolved;
    bool outcome;
    bool callbackOnValidate;
    bytes callbackData;
}

contract ValidationPools is Reputation, Forum {
    mapping(uint => ValidationPool) public validationPools;
    uint public validationPoolCount;

    mapping(uint => address) public members;
    uint public memberCount;
    mapping(address => bool) public isMember;

    uint constant minDuration = 1; // 1 second
    uint constant maxDuration = 365_000_000 days; // 1 million years
    uint[2] minQuorum = [1, 10];

    event ValidationPoolInitiated(uint poolIndex);
    event ValidationPoolResolved(
        uint poolIndex,
        bool votePasses,
        bool quorumMet
    );

    /// Internal function to register a stake for/against a validation pool
    function _stakeOnValidationPool(
        ValidationPool storage pool,
        address sender,
        uint256 amount,
        bool inFavor
    ) internal {
        require(block.timestamp <= pool.endTime, "Pool end time has passed");
        // We don't call _update here; We defer that until evaluateOutcome.
        uint stakeIndex = pool.stakeCount++;
        ValidationPoolStake storage s = pool.stakes[stakeIndex];
        s.sender = sender;
        s.inFavor = inFavor;
        s.amount = amount;
        s.id = stakeIndex;
    }

    /// Accept reputation stakes toward a validation pool
    function stakeOnValidationPool(
        uint poolIndex,
        uint256 amount,
        bool inFavor
    ) public {
        ValidationPool storage pool = validationPools[poolIndex];
        _stakeOnValidationPool(pool, msg.sender, amount, inFavor);
    }

    /// Accept reputation stakes toward a validation pool
    function delegatedStakeOnValidationPool(
        uint poolIndex,
        address owner,
        uint256 amount,
        bool inFavor
    ) public {
        ValidationPool storage pool = validationPools[poolIndex];
        _spendAllowance(owner, msg.sender, amount);
        _stakeOnValidationPool(pool, owner, amount, inFavor);
    }

    /// Accept fee to initiate a validation pool
    function initiateValidationPool(
        uint postIndex,
        uint duration,
        uint[2] calldata quorum, // [Numerator, Denominator]
        uint[2] calldata winRatio, // [Numerator, Denominator]
        uint bindingPercent,
        bool redistributeLosingStakes,
        bool callbackOnValidate,
        bytes calldata callbackData
    ) external payable returns (uint poolIndex) {
        require(msg.value > 0, "Fee is required to initiate validation pool");
        require(duration >= minDuration, "Duration is too short");
        require(duration <= maxDuration, "Duration is too long");
        require(
            minQuorum[1] * quorum[0] >= minQuorum[0] * quorum[1],
            "Quorum is below minimum"
        );
        require(quorum[0] <= quorum[1], "Quorum is greater than one");
        require(winRatio[0] <= winRatio[1], "Win ratio is greater than one");
        require(bindingPercent <= 100, "Binding percent must be <= 100");
        Post storage post = posts[postIndex];
        require(post.author != address(0), "Target post not found");
        poolIndex = validationPoolCount++;
        ValidationPool storage pool = validationPools[poolIndex];
        pool.sender = msg.sender;
        pool.postIndex = postIndex;
        pool.fee = msg.value;
        pool.params.quorum = quorum;
        pool.params.winRatio = winRatio;
        pool.params.bindingPercent = bindingPercent;
        pool.params.redistributeLosingStakes = redistributeLosingStakes;
        pool.params.duration = duration;
        pool.endTime = block.timestamp + duration;
        pool.id = poolIndex;
        pool.callbackOnValidate = callbackOnValidate;
        pool.callbackData = callbackData;
        // We use our privilege as the DAO contract to mint reputation in proportion with the fee.
        // Here we assume a minting ratio of 1
        // TODO: Make minting ratio an adjustable parameter
        _mint(address(this), msg.value);
        pool.minted = msg.value;
        emit ValidationPoolInitiated(poolIndex);
    }

    /// Evaluate outcome of a validation pool
    function evaluateOutcome(uint poolIndex) public returns (bool votePasses) {
        ValidationPool storage pool = validationPools[poolIndex];
        Post storage post = posts[pool.postIndex];
        require(pool.resolved == false, "Pool is already resolved");
        uint stakedFor;
        uint stakedAgainst;
        ValidationPoolStake storage s;
        for (uint i = 0; i < pool.stakeCount; i++) {
            s = pool.stakes[i];
            if (s.inFavor) {
                stakedFor += s.amount;
            } else {
                stakedAgainst += s.amount;
            }
        }
        stakedFor += pool.minted / 2;
        stakedAgainst += pool.minted / 2;
        if (pool.minted % 2 != 0) {
            stakedFor += 1;
        }
        // Special case for early evaluation if dao.totalSupply has been staked
        require(
            block.timestamp > pool.endTime ||
                stakedFor + stakedAgainst == totalSupply(),
            "Pool end time has not yet arrived"
        );
        // Check that quorum is met
        if (
            pool.params.quorum[1] * (stakedFor + stakedAgainst) <=
            totalSupply() * pool.params.quorum[0]
        ) {
            // TODO: Refund fee
            // TODO: this could be made available for the sender to withdraw
            // payable(pool.sender).transfer(pool.fee);
            pool.resolved = true;
            emit ValidationPoolResolved(poolIndex, false, false);
            // Callback if requested
            if (pool.callbackOnValidate) {
                IOnValidate(pool.sender).onValidate(
                    votePasses,
                    false,
                    stakedFor,
                    stakedAgainst,
                    pool.callbackData
                );
            }

            return false;
        }
        // A tie is resolved in favor of the validation pool.
        // This is especially important so that the DAO's first pool can pass,
        // when no reputation has yet been minted.

        votePasses =
            stakedFor * pool.params.winRatio[1] >=
            (stakedFor + stakedAgainst) * pool.params.winRatio[0];
        if (votePasses && !isMember[post.author]) {
            members[memberCount++] = post.author;
            isMember[post.author] = true;
        }
        pool.resolved = true;
        pool.outcome = votePasses;
        emit ValidationPoolResolved(poolIndex, votePasses, true);

        // Value of losing stakes should be distributed among winners, in proportion to their stakes
        // Only bindingPercent % should be redistributed
        // Stake senders should get (100-bindingPercent) % back
        uint amountFromWinners = votePasses ? stakedFor : stakedAgainst;
        uint totalRewards;
        uint totalAllocated;
        for (uint i = 0; i < pool.stakeCount; i++) {
            s = pool.stakes[i];
            if (votePasses != s.inFavor) {
                // Losing stake
                uint amount = (s.amount * pool.params.bindingPercent) / 100;
                if (pool.params.redistributeLosingStakes) {
                    _update(s.sender, address(this), amount);
                    totalRewards += amount;
                } else {
                    _burn(s.sender, amount);
                }
            }
        }

        if (votePasses) {
            // If vote passes, reward the author as though they had staked the winning portion of the VP initial stake
            // Here we assume a stakeForAuthor ratio of 0.5
            // TODO: Make stakeForAuthor an adjustable parameter
            totalRewards += pool.minted / 2;
            // Include the losign portion of the VP initial stake
            // Issue rewards to the winners
            for (uint i = 0; i < pool.stakeCount; i++) {
                s = pool.stakes[i];
                if (
                    pool.params.redistributeLosingStakes &&
                    votePasses == s.inFavor
                ) {
                    // Winning stake
                    uint reward = (((totalRewards * s.amount) /
                        amountFromWinners) * pool.params.bindingPercent) / 100;
                    totalAllocated += reward;
                    _update(address(this), s.sender, reward);
                }
            }
            // Due to rounding, there may be some excess REP. Award it to the author.
            uint remainder = totalRewards - totalAllocated;
            if (pool.minted % 2 != 0) {
                // We staked the odd remainder in favor of the post, on behalf of the author.
                remainder += 1;
            }

            // Transfer REP to the forum instead of to the author directly
            _onValidatePost(pool.postIndex, pool.minted / 2 + remainder);
        } else {
            // If vote does not pass, divide the losing stake among the winners
            totalRewards += pool.minted;
            for (uint i = 0; i < pool.stakeCount; i++) {
                s = pool.stakes[i];
                if (
                    pool.params.redistributeLosingStakes &&
                    votePasses == s.inFavor
                ) {
                    // Winning stake
                    uint reward = (((totalRewards * s.amount) /
                        (amountFromWinners - pool.minted / 2)) *
                        pool.params.bindingPercent) / 100;
                    totalAllocated += reward;
                    _update(address(this), s.sender, reward);
                }
            }
        }

        // Distribute fee proportionately among all reputation holders
        for (uint i = 0; i < memberCount; i++) {
            address member = members[i];
            uint share = (pool.fee * balanceOf(member)) / totalSupply();
            // TODO: For efficiency this could be modified to hold the funds for recipients to withdraw
            payable(member).transfer(share);
        }

        // Callback if requested
        if (pool.callbackOnValidate) {
            IOnValidate(pool.sender).onValidate(
                votePasses,
                true,
                stakedFor,
                stakedAgainst,
                pool.callbackData
            );
        }
    }
}
